#include <fstream>
#include <opencv2/opencv.hpp>
#include "GCoptimization.h"

class LightField
{
  public:
    int height;
    int width;
    int num_u;
    int num_v;
    int num_label;
    std::string path;
    std::vector<std::vector<cv::Mat> > LF;
    cv::Mat dispImg;

    std::vector<cv::Mat> seed;
    std::vector<std::vector<std::vector<cv::Mat> > > likelihood_image;
    std::vector<cv::Mat> seed_exists;
    std::vector<cv::Mat> hist;
    float size_bin;
    int num_bin;
    std::vector<cv::Mat> likelihood;

    double k1;
    double k2;
    double k3;
    double K;

    LightField() :height(0), width(0), num_u(0), num_v(0), num_label(2), size_bin(0.f), num_bin(0), path(""), k1(1), k2(1), k3(1), K(1){};

    static void load_params(const std::string &filename, std::map<std::string, std::string> &paramList);
    void load_img(const int _height, const int _width, const int _numU, const int _numV, const int _numBin, const std::string &_path, const int labelNum);
    static void seed2hist(const cv::Mat &_seed, const cv::Mat &_isExist, const int _bins, cv::Mat &_hist, cv::Mat &_likelihood, const int _epsilon = 0);
    inline void SiteID2LFID(const int _siteID, int &_x, int &_y, int &_u, int &_v) const {
      int s = _siteID;
      _x = s%width;     s = s / width;
      _y = s%height;    s = s / height;
      _u = s%num_u;     s = s / num_u;
      _v = s/*%num_v*/;
    };
    inline void LFID2SiteID(int &_siteID, const int _x, const int _y, const int _u, const int _v) const {
      _siteID = ((((_v*num_u) + _u)*height) + _y)*width + _x;
    };

};

void LightField::load_params(const std::string &filename, std::map<std::string, std::string> &paramList)
{
  std::map<std::string, std::string> list;

  std::ifstream fi(filename.c_str());
  while (fi) {
    std::string line;
    std::getline(fi, line);
    if (!fi) break;

    unsigned int c = 0;
    while (line[c] != ' ' && line.length() > c) c++;
    std::string key = line.substr(0, c);

    c = line.length();
    while (line[c - 1] != ' ' && c > 0) c--;
    std::string val_s = line.substr(c, line.length() - c);

    list[key] = val_s;
  }
  paramList = list;
}

void LightField::load_img(const int _height, const int _width, const int _numU, const int _numV, const int _numBin, const std::string &_path, const int labelNum)
{
  std::vector<std::vector<cv::Mat> > _LF;
  height = _height;
  width = _width;
  num_u = _numU;
  num_v = _numV;
  path = _path;
  num_label = labelNum;

  num_bin = _numBin;
  size_bin = 256.f / _numBin;

  _LF.resize(num_v);
  for (int v = 0; v < num_v; v++) {
    _LF[v].resize(num_u);
    for (int u = 0; u < num_u; u++) {
      char imgPath[1024];
      sprintf(imgPath, "%s%d_%d.png", path.c_str(), v, u);

      cv::Mat img = cv::imread(std::string(imgPath)); 
         

      if (img.empty()) {
        throw std::string("Cannot load ") + std::string(imgPath);
      }
      else {
        std::cerr << imgPath << " has successfully loaded." << std::endl;
        cv::cvtColor(img, _LF[v][u], cv::COLOR_BGR2YCrCb);
      }
    }
  }

  seed.resize(num_label);
  seed_exists.resize(num_label);

  //std::vector<std::vector<std::vector<cv::Mat> > > likelihood_image;

  likelihood_image.resize(num_label);
  
  hist.resize(num_label);
  likelihood.resize(num_label);

  //cv::namedWindow("win");
  for (int l = 0; l < num_label; l++) {
    
    char filename[1024];

    sprintf(filename, "seed%d.png", l);
    seed[l] = cv::imread(path + std::string(filename), -1);   // 0th to 2nd has same value and 3 has alpha value.

    likelihood_image[l].resize(num_v);
    for (int v = 0; v < num_v; v++){
      likelihood_image[l][v].resize(num_u);
      for (int u = 0; u < num_u; u++){
        
        char likelihoodname[1024];
        sprintf(likelihoodname, "likelihood/%d_%d_%d.png", l, v, u);
        std::cout << likelihoodname << std::endl;
        likelihood_image[l][v][u] = cv::imread(path + std::string(likelihoodname), 0); //read likelihood image
        
        
        //if (seed[l].empty())    throw "Cannot load seed";
        if ( likelihood_image[l][v][u].empty()){
          throw "Cannot load likelihood_image";
        }else{
          //std::cerr <<  likelihood_image[0][0][0] << std::endl;
        }
      }
    }



    std::vector<cv::Mat> channels;
    cv::split(seed[l], channels);
    seed_exists[l] = channels[3].clone();
    // BGR to YCR_CB
    cv::cvtColor(seed[l], seed[l], cv::COLOR_BGR2YCrCb);
    // calculate histogram
    seed2hist(seed[l], seed_exists[l], num_bin, hist[l], likelihood[l], 1);
  }
  std::cout << likelihood_image.size() << std::endl;

  dispImg = cv::imread(path + std::string("disparity.png"), 0) / 10.;

  LF.swap(_LF);

  return;
}

void LightField::seed2hist(const cv::Mat &_seed, const cv::Mat &_isExist, const int _bins, cv::Mat &_hist, cv::Mat &_likelihood, const int _epsilon)
{
  int histSize[] = { _bins, _bins };
  float range[] = { 0, 256 };
  const float* ranges[] = { range, range, };
  int channels[] = { 1, 2 };

  cv::calcHist(&_seed, 1, (int*)channels, _isExist, _hist, 2, (const int*)histSize, (const float**)ranges, true, false);


  _likelihood = _hist + _epsilon;
  cv::normalize(_likelihood, _likelihood, 1, 0, cv::NORM_L1);
  cv::log(_likelihood, _likelihood);
  _likelihood = _likelihood*-1;


  return;
}

//static const GCoptimization::EnergyTermType MaxWeight = 100;

GCoptimization::EnergyTermType calcDataTerm(GCoptimization::SiteID s, GCoptimization::LabelID l, void *extraData)
{

  LightField* pLF = (LightField*)extraData;

  const int ctr_u = (pLF->num_u / 2);
  const int ctr_v = (pLF->num_v / 2);

  int x, y, u, v;
  pLF->SiteID2LFID(s, x, y, u, v);

  /*
  if (u == ctr_u && v == ctr_v) {
    bool checkSeed = false;
    for (int k = 0; k < pLF->num_label; k++) {
      checkSeed |= (pLF->seed_exists[k].at<uchar>(y, x) != 0);
    }
  }
  */
  

  //cv::Vec3b YCbCr = pLF->LF[v][u].at<cv::Vec3b>(y, x);

  //GCoptimization::EnergyTermType weight = (GCoptimization::EnergyTermType)(pLF->k1 * pLF->likelihood_image[l][v][u].at<unsigned char>(y, x) / (-100.) + 100.);
  
  if ( pLF->likelihood_image[l][v][u].empty()){
    throw "Cannot load likelihood_image";
  }else{
    //std::cerr << pLF->likelihood_image[0][0][0] << std::endl;
  }


  GCoptimization::EnergyTermType weight = (GCoptimization::EnergyTermType)(pLF->k1*pLF->likelihood_image[int(l)][v][u].at<unsigned char>(y, x) / (-100.) + 1.);

  return weight;
}

GCoptimization::EnergyTermType calcSmoothTerm(GCoptimization::SiteID s1, GCoptimization::SiteID s2, GCoptimization::LabelID l1, GCoptimization::LabelID l2, void *extraData)
{

  if (l1 == l2) return 0;
  LightField* pLF = (LightField*)extraData;

  int x1, y1, u1, v1, x2, y2, u2, v2;
  pLF->SiteID2LFID(s1, x1, y1, u1, v1);
  pLF->SiteID2LFID(s2, x2, y2, u2, v2);

  // calc weight
  double weight = 1;
  {
    const double kappa_const = 1.;
    const double lambda_const = 1.;
    double dist = 1;
    double tmp = -1. * kappa_const;

    cv::Vec3b p1 = pLF->LF[v1][u1].at<cv::Vec3b>(y1, x1),
      p2 = pLF->LF[v2][u2].at<cv::Vec3b>(y2, x2);

    double l2norm = 0;
    for (int n = 0; n < 3; n++) { double d = (double)(p1.val[n] - p2.val[n]) / 255.; l2norm += d*d; }
    weight = lambda_const * exp(tmp*l2norm) / dist;
  }

  bool term2 = (u1 == u2 && v1 == v2);

  if (term2) {
    return std::max<GCoptimization::EnergyTermType>(0, pLF->k2*weight);
  }
  else {
    // term3
    return std::max<GCoptimization::EnergyTermType>(0, pLF->k3*weight);
  }
}


void draw_result(GCoptimizationGeneralGraph &_gcogg, const LightField &_LF, int p1, int p2)
{
  for (int l = 0; l < _LF.num_label; l++) {
    for (int v = 0; v < _LF.num_v; v++) {
      for (int u = 0; u < _LF.num_u; u++) {
        cv::Mat res;
        cv::cvtColor(_LF.LF[v][u], res, cv::COLOR_YCrCb2BGR);

        std::vector<cv::Mat> channels;
        cv::split(res, channels);

        channels.push_back(channels[0].clone()*0);  // alpha channel

        for (int y = 0; y < _LF.height; y++) {
          for (int x = 0; x < _LF.width; x++) {
            int SiteID;
            _LF.LFID2SiteID(SiteID, x, y, u, v);

            channels[3].at<uchar>(y, x) = (_gcogg.whatLabel(SiteID) == l ? 255 : 0);
            
          }
        }
        cv::merge(channels, res);

        char filepath[1024];
        sprintf(filepath, "./out/%d/%d_%d.png", l, v, u);

        //cv::namedWindow("win");
        //cv::imshow("win", res);
        //cv::waitKey(50);  
        
        cv::imwrite(filepath, res);
      }
    }
  }

    for (int v = 0; v < _LF.num_v; v++) {
      for (int u = 0; u < _LF.num_u; u++) {
        cv::Mat res(_LF.height,_LF.width,CV_8U);
        for (int y = 0; y < _LF.height; y++) {
          for (int x = 0; x < _LF.width; x++) {
            int SiteID;
            _LF.LFID2SiteID(SiteID, x, y, u, v);
            res.at<uchar>(y, x) = _gcogg.whatLabel(SiteID);
          }
        }
        char filepath[1024];
        sprintf(filepath, "./out/%d_%d.png", v, u);
        cv::imwrite(filepath, res);
      }
    }
}

int main(int argc, char* argv[])
{
  /*
  if (argc != 2) {
    std::cerr << "Usage: python LFSegmenter.py <path>" << std::endl;
    exit(-1);
  }
  */

  std::string path = argv[1];
  std::map<std::string, std::string> params;
  std::cout << path << std::endl;
  LightField::load_params(path + std::string("script.txt"), params);
  LightField LF;

  LF.k1 = atof(params["k1"].c_str()) * 10;
  LF.k2 = atof(params["k2"].c_str()) * 10;
  LF.k3 = atof(params["k3"].c_str()) * 10;


  std::cout << LF.k1 << std::endl;
  std::cout << LF.k2 << std::endl;
  std::cout << LF.k3 << std::endl;

  LF.K = LF.k1;


  //LF.num_label = 4;

  int num_u = atoi(params["num_u"].c_str());
  int num_v = atoi(params["num_v"].c_str());
  int height = atoi(params["height"].c_str());
  int width = atoi(params["width"].c_str());
  int binsize = atoi(params["binsize"].c_str());

  LF.num_label = atof(params["num_label"].c_str());

  const int labelNum = LF.num_label;

  LF.load_img(height, width, num_u, num_v, binsize, path, labelNum);
  std::cerr << "load img" << std::endl;


  const int SiteNum = height*width*num_u*num_v;


  GCoptimizationGeneralGraph gcogg(SiteNum, labelNum);

  const int MaxNeighbors = 4 + num_u*num_v;
  //const int MaxNeighbors = 4 + 4;
  GCoptimization::SiteID *numNeighbors = new GCoptimization::SiteID[SiteNum];
  GCoptimization::SiteID **neighborsIndexes = new GCoptimization::SiteID*[SiteNum];
  GCoptimization::SiteID *neighborsIndexesBuf = new GCoptimization::SiteID[SiteNum*MaxNeighbors];
  GCoptimization::EnergyTermType **neighborsWeights = new GCoptimization::EnergyTermType*[SiteNum];
  GCoptimization::EnergyTermType fixedWeight[MaxNeighbors];
  for ( int i=0; i < MaxNeighbors; i++ ) fixedWeight[i] = 1;
  
  const int ctr_u = (num_u / 2);
  const int ctr_v = (num_v / 2);

  for (int v = 0; v < num_v; v++) {
    for (int u = 0; u < num_u; u++) {
      for (int y = 0; y < height; y++) {
        for (int x = 0; x < width; x++) {
          int SiteNum = 0;
          LF.LFID2SiteID(SiteNum, x, y, u, v);
          numNeighbors[SiteNum] = 0;
          neighborsIndexes[SiteNum] = NULL;
          neighborsWeights[SiteNum] = fixedWeight;

          std::vector<GCoptimization::SiteID> neighbor(MaxNeighbors);
          int count = 0;

          // term2 - Neighbour
          if (x != 0)      LF.LFID2SiteID(neighbor[count++], x - 1, y, u, v);
          if (y != 0)      LF.LFID2SiteID(neighbor[count++], x, y - 1, u, v);
          if (x != width - 1)   LF.LFID2SiteID(neighbor[count++], x + 1, y, u, v);
          if (y != height - 1)  LF.LFID2SiteID(neighbor[count++], x, y + 1, u, v);

          // term3

          if (u != ctr_u || v != ctr_v) { // if not center

            int new_x = 1.*LF.dispImg.at<uchar>(y, x) * (u - ctr_u) + x,
                new_y = 1.*LF.dispImg.at<uchar>(y, x) * (v - ctr_v) + y;

            if (new_x >= 0 && new_x < width && new_y >= 0 && new_y < height) {
              LF.LFID2SiteID(neighbor[count++], new_x, new_y, ctr_u, ctr_v);
            }
          }
          else {   // if center
            for (int dv = 0; dv < num_v; dv++) {
              if (dv == ctr_v) continue;
              for (int du = 0; du < num_u; du++) {
                if (du == ctr_u) continue;

                int new_x = 1.*LF.dispImg.at<uchar>(y, x) /* / 30. */ * (du - ctr_u) - x,
                    new_y = 1.*LF.dispImg.at<uchar>(y, x) /* / 30. */ * (dv - ctr_v) - y;

                if (new_x >= 0 && new_x < width && new_y >= 0 && new_y < height) {
                  LF.LFID2SiteID(neighbor[count++], new_x, new_y, du, dv);
                }
              }
            }
          }


          assert(count <= MaxNeighbors);

          numNeighbors[SiteNum] = count;
          neighborsIndexes[SiteNum] = &neighborsIndexesBuf[SiteNum*MaxNeighbors];
          for (int n = 0; n < numNeighbors[SiteNum]; n++) {
            neighborsIndexes[SiteNum][n] = neighbor[n];
          }
        }
      }
    }
  }
  gcogg.setAllNeighbors(numNeighbors, neighborsIndexes, neighborsWeights);

  gcogg.setDataCost(calcDataTerm, &LF);
  gcogg.setSmoothCost(calcSmoothTerm, &LF);

  std::cerr << "start optimization" << std::endl;
  gcogg.swap();
  std::cerr << "optimization complete" << std::endl;

  draw_result(gcogg, LF, LF.k2, LF.k3);

  delete[] numNeighbors;
  delete[] neighborsIndexes;
  delete[] neighborsIndexesBuf;
  delete[] neighborsWeights;

  std::cerr << "end" << std::endl;

  return 0;
}
